
/* =========================================================================== *
|
|  Copyright (c) 1994-2008 by Kobus Barnard (author).
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowledged in publications, and relevant papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarantee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#ifndef L_ERROR_INCLUDED
#define L_ERROR_INCLUDED

#include "l/l_def.h"

#ifdef __cplusplus
extern "C" {
#ifdef COMPILING_CPLUSPLUS_SOURCE
namespace kjb_c {
#endif
#endif

/*
// Lint complains about null clauses. Thus we give up and define two versions.
// One double one, and one for lint. The null clauses are not that critical,
// but they gaurd against a certain kind of error--unfortunately, I can't
// remember what that error is! Nonetheless, keep it for now.
*/

#ifndef __lint

/* =============================================================================
 *                             NOTE_ERROR
 *
 * (MACRO) Adds a line to the error message store
 *
 * If the debug level is greater than 0, this macro addds an error to the list
 * of lines stored in the current error list with add_error(); The line is
 * simply:
 * |    Error noted on line [line] of [file].
 *
 *
 * Index: debugging, error handling
 *
 * -----------------------------------------------------------------------------
*/
#ifdef __C2MAN__
    void NOTE_ERROR(void);
#else

#define NOTE_ERROR() \
             {                                                         \
                  /* extern int kjb_debug_level; */                    \
                                                                       \
                  if (kjb_debug_level > 0)                             \
                  {                                                    \
                      add_error("Error noted on line %d of %s.",       \
                                __LINE__, __FILE__);                   \
                  }                                                    \
              }
#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             ERE
 *
 * (MACRO) Returns ERROR if its argument is ERROR.
 *
 * This macro becomes an ERROR return if its argument is ERROR. It is a
 * convenient quick return that makes sense in the KJB library, but be careful
 * about cleanup! Often code with ERE()'s will leak memory on failure which is
 * acceptiable if the failure is rare, but this should be documented. 
 *
 * Further, if the debug level is greater than 0, then a message is added to the
 * list of messages using add_errror(); That line is:
 * |    (ERE on line [line] of [file].
 *
 *
 * Index: debugging, error handling
 *
 * -----------------------------------------------------------------------------
*/
#ifdef __C2MAN__
    int ERE(int);
#else

#define ERE(x)        if ((x) == ERROR)                                        \
                      {                                                        \
                          /* extern int kjb_debug_level; */                    \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(ERE on line %d of %s.)",             \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          return ERROR;                                        \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             ERN
 *
 * (MACRO) Returns NULL if its argument is ERROR.
 *
 * This macro becomes a NULL return if its argument is ERROR. It is a convenient
 * quick return that makes sense in the KJB library, but be careful about
 * cleanup! Often code with ERN()'s will leak memory on failure which is
 * acceptiable if the failure is rare, but this should be documented. 
 *
 * Further, if the debug level is greater than 0, then a message is added to the
 * list of messages using add_errror(); That line is:
 * |    (ERN on line [line] of [file].
 *
 *
 * Index: debugging, error handling
 *
 * -----------------------------------------------------------------------------
*/
#ifdef __C2MAN__
    void* ERN(int);
#else

#define ERN(x)        if ((x) == ERROR)                                        \
                      {                                                        \
                          /* extern int kjb_debug_level; */                    \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(ERN on line %d of %s.)",             \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          return NULL;                                         \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             ER
 *
 * (MACRO) Returns (void) if its argument is ERROR.
 *
 * This macro becomes a simple return if its argument is ERROR. It is a
 * convenient quick return that makes sense in the KJB library, but be careful
 * about cleanup! Often code with ERE()'s will leak memory on failure which is
 * acceptiable if the failure is rare, but this should be documented. 
 *
 * Further, if the debug level is greater than 0, then a message is added to the
 * list of messages using add_errror(); That line is:
 * |    (ER on line [line] of [file].
 *
 *
 * Index: debugging, error handling
 *
 * -----------------------------------------------------------------------------
*/
#ifdef __C2MAN__
    void ER(int);
#else

#define ER(x)         if ((x) == ERROR)                                        \
                      {                                                        \
                          /* extern int kjb_debug_level; */                    \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(ER on line %d of %s.)",              \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          return;                                              \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }
#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             EPE
 *
 * (MACRO) Prints ERROR if its argument is ERROR.
 *
 * This macro prints the if its argument for ERROR.  
 *
 * Further, if the debug level is greater than 0, then a message is added to the
 * list of messages using add_errror(); That line is:
 * |    (EPE on line [line] of [file].
 *
 *
 * Index: debugging, error handling
 *
 * -----------------------------------------------------------------------------
*/
#ifdef __C2MAN__
    void EPE(int);
#else

#define EPE(x)        if ((x) == ERROR)                                        \
                      {                                                        \
                          /* extern int kjb_debug_level; */                    \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(EPE on line %d of %s.)",             \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          kjb_print_error();                                   \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }
#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             ETE
 *
 * (MACRO) Terminates execution if its argument is ERROR.
 *
 * This macro terminates execution if its argument for ERROR. More often, one
 * wants to also print the error message which is doen with EPETE();
 *
 * Further, if the debug level is greater than 0, then a message is added to the
 * list of messages using add_errror(); That line is:
 * |    (ETE on line [line] of [file].
 *
 *
 * Index: debugging, error handling
 *
 * -----------------------------------------------------------------------------
*/
#ifdef __C2MAN__
    void ETE(int);
#else

#define ETE(x)        if ((x) == ERROR)                                        \
                      {                                                        \
                          /* extern int kjb_debug_level; */                    \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(ETE on line %d of %s.)",             \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          kjb_exit(EXIT_FAILURE);                              \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             EPETE
 *
 * (MACRO) Prints error and terminates execution if its argument is ERROR.
 *
 * This macro prints the error message and terminates execution if its argument
 * for ERROR. 
 *
 * Further, if the debug level is greater than 0, then a message is added to the
 * list of messages using add_errror(); That line is:
 * |    (EPETE on line [line] of [file].
 *
 *
 * Index: debugging, error handling
 *
 * -----------------------------------------------------------------------------
*/
#ifdef __C2MAN__
    void EPETE(int);
#else

#define EPETE(x)      if ((x) == ERROR)                                        \
                      {                                                        \
                          /* extern int kjb_debug_level; */                    \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(EPETE on line %d of %s.)",           \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          kjb_print_error();                                   \
                          kjb_exit(EXIT_FAILURE);                              \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#define ESBRE(x)      if ((x) == ERROR)                                        \
                      {                                                        \
                          kjb_print_error();                                   \
                          set_bug("Treating error return on line %d of %s as a bug.)",  \
                                        __LINE__, __FILE__);                   \
                          return ERROR;                                        \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#define BATCH_EPETE(x) \
                      if ((x) == ERROR)                                        \
                      {                                                        \
                          /* extern int kjb_debug_level; */                    \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(BATCH_EPETE on line %d of %s.)",     \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          kjb_print_error();                                   \
                                                                               \
                          if ( ! is_interactive())                             \
                          {                                                    \
                              kjb_exit(EXIT_FAILURE);                          \
                          }                                                    \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             NRE
 *
 * (MACRO) Returns NULL if its argument is ERROR.
 *
 * This macro becomes an NULL return if its argument is ERROR. It is a
 * convenient quick return that makes sense in the KJB library, but be careful
 * about cleanup! Often code with NRE()'s will leak memory on failure which is
 * acceptiable if the failure is rare, but this should be documented. 
 *
 * Further, if the debug level is greater than 0, then a message is added to the
 * list of messages using add_errror(); That line is:
 * |    (NRE on line [line] of [file].
 *
 *
 * Index: debugging, error handling
 *
 * -----------------------------------------------------------------------------
*/
#ifdef __C2MAN__
    int NRE(int);
#else

#define NRE(x)        if ((x) == NULL)                                         \
                      {                                                        \
                          /* extern int kjb_debug_level; */                    \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(NRE on line %d of %s.)",             \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          return ERROR;                                        \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             NRN
 *
 * (MACRO) Returns NULL if its argument is NULL.
 *
 * This macro becomes a NULL return if its argument is ERROR. It is a convenient
 * quick return that makes sense in the KJB library, but be careful about
 * cleanup! Often code with NRN()'s will leak memory on failure which is
 * acceptiable if the failure is rare, but this should be documented. 
 *
 * Further, if the debug level is greater than 0, then a message is added to the
 * list of messages using add_errror(); That line is:
 * |    (NRN on line [line] of [file].
 *
 *
 * Index: debugging, error handling
 *
 * -----------------------------------------------------------------------------
*/
#ifdef __C2MAN__
    void* NRN(int);
#else

#define NRN(x)        if ((x) == NULL)                                         \
                      {                                                        \
                          /* extern int kjb_debug_level; */                    \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(NRN on line %d of %s.)",             \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          return NULL;                                         \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }
#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             NR
 *
 * (MACRO) Returns (void) if its argument is NULL.
 *
 * This macro becomes a simple return if its argument is NULL. It is a
 * convenient quick return that makes sense in the KJB library, but be careful
 * about cleanup! Often code with NR()'s will leak memory on failure which is
 * acceptiable if the failure is rare, but this should be documented. 
 *
 * Further, if the debug level is greater than 0, then a message is added to the
 * list of messages using add_errror(); That line is:
 * |    (NR on line [line] of [file].
 *
 *
 * Index: debugging, error handling
 *
 * -----------------------------------------------------------------------------
*/
#ifdef __C2MAN__
    void NR(int);
#else

#define NR(x)         if ((x) == NULL)                                         \
                      {                                                        \
                          /* extern int kjb_debug_level; */                    \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(NR on line %d of %s.)",              \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          return;                                              \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             NPE
 *
 * (MACRO) Prints ERROR if its argument is NULL.
 *
 * This macro prints the if its argument for NULL.  
 *
 * Further, if the debug level is greater than 0, then a message is added to the
 * list of messages using add_errror(); That line is:
 * |    (NPE on line [line] of [file].
 *
 *
 * Index: debugging, error handling
 *
 * -----------------------------------------------------------------------------
*/
#ifdef __C2MAN__
    void NPE(void*);
#else

#define NPE(x)        if ((x) == NULL)                                         \
                      {                                                        \
                          /* extern int kjb_debug_level; */                    \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(NPE on line %d of %s.)",             \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          kjb_print_error();                                   \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             NTE
 *
 * (MACRO) Terminates execution if its argument is NULL.
 *
 * This macro terminates execution if its argument for ERROR. More often, one
 * wants to also print the error message which is doen with NPETE();
 *
 * Further, if the debug level is greater than 0, then a message is added to the
 * list of messages using add_errror(); That line is:
 * |    (NTE on line [line] of [file].
 *
 *
 * Index: debugging, error handling
 *
 * -----------------------------------------------------------------------------
*/
#ifdef __C2MAN__
    void NTE(int);
#else

#define NTE(x)        if ((x) == NULL)                                         \
                      {                                                        \
                          /* extern int kjb_debug_level; */                    \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(NTE on line %d of %s.)",             \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          kjb_exit(EXIT_FAILURE);                              \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             NPETE
 *
 * (MACRO) Prints error and terminates execution if its argument is NULL.
 *
 * This macro prints the error message and terminates execution if its argument
 * for NULL. 
 *
 * Further, if the debug level is greater than 0, then a message is added to the
 * list of messages using add_errror(); That line is:
 * |    (NPETE on line [line] of [file].
 *
 *
 * Index: debugging, error handling
 *
 * -----------------------------------------------------------------------------
*/
#ifdef __C2MAN__
    void NPETE(void*);
#else

#define NPETE(x)      if ((x) == NULL)                                         \
                      {                                                        \
                          /* extern int kjb_debug_level; */                    \
                                                                               \
                          if (kjb_debug_level > 0)                             \
                          {                                                    \
                              add_error("(NPETE on line %d of %s.)",           \
                                        __LINE__, __FILE__);                   \
                          }                                                    \
                          kjb_print_error();                                   \
                          kjb_exit(EXIT_FAILURE);                              \
                      }                                                        \
                      else                                                     \
                      {                                                        \
                          ; /* Do nothing */                                   \
                      }
#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#else

/*
--------------------------------------------------------------------------------
|                       Definitions for "lint" only
|
|                                  ||
|                                 \||/
|                                  \/
*/

#define ERE(x)        if ((x) == ERROR)                                        \
                      {                                                        \
    /* LINT */            return ERROR;                                        \
                      }

#define ERN(x)        if ((x) == ERROR)                                        \
                      {                                                        \
    /* LINT */            return NULL;                                         \
                      }

#define ER(x)         if ((x) == ERROR)                                        \
                      {                                                        \
                          return;                                              \
    /* LINT */        }

#define EPE(x)        if ((x) == ERROR)                                        \
                      {                                                        \
    /* LINT */            kjb_print_error();                                   \
                      }

#define ETE(x)        if ((x) == ERROR)                                        \
                      {                                                        \
    /* LINT */            kjb_exit(EXIT_FAILURE);                              \
                      }

#define EPETE(x)      if ((x) == ERROR)                                        \
                      {                                                        \
                          kjb_print_error();                                   \
    /* LINT */            kjb_exit(EXIT_FAILURE);                              \
                      }

#define ESBRE(x)      if ((x) == ERROR)                                        \
                      {                                                        \
                          kjb_print_error();                                   \
                          set_bug("Treating error returnas a bug.)");          \
    /* LINT */            return ERROR;                                        \
                      }

#define BATCH_EPETE(x)    if ((x) == ERROR)                                    \
                          {                                                    \
                              kjb_print_error();                               \
    /* LINT */                kjb_exit(EXIT_FAILURE);                          \
                          }

#define NRE(x)        if ((x) == NULL)                                         \
                      {                                                        \
                          return ERROR;                                        \
    /* LINT */        }

#define NRN(x)        if ((x) == NULL)                                         \
                      {                                                        \
    /* LINT */            return NULL;                                         \
                      }

#define NR(x)         if ((x) == NULL)                                         \
                      {                                                        \
                          return;                                              \
    /* LINT */        }

#define NPE(x)        if ((x) == NULL)                                         \
                      {                                                        \
    /* LINT */            kjb_print_error();                                   \
                      }

#define NTE(x)        if ((x) == NULL)                                         \
                      {                                                        \
    /* LINT */            kjb_exit(EXIT_FAILURE);                              \
                      }

#define NPETE(x)      if ((x) == NULL)                                         \
                      {                                                        \
    /* LINT */            kjb_print_error();                                   \
                          kjb_exit(EXIT_FAILURE);                              \
                      }
/*
|                                  /\
|                                 /||\
|                                  ||
|
|                       Definitions for "lint" only
--------------------------------------------------------------------------------
*/

#endif

void check_num_args(int, int, int, const char*);

#ifdef __cplusplus
#ifdef COMPILING_CPLUSPLUS_SOURCE
}
#endif
}
#endif

#endif

