
/* =========================================================================== *
|
|  Copyright (c) 1994-2008 by Kobus Barnard (author).
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowledged in publications, and relevant papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarantee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "n/n_cholesky.h"

#ifdef __cplusplus
extern "C" {
#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* -------------------------------------------------------------------------- */

/* =============================================================================
 *                          cholesky_decomposition
 *
 * Performs Cholesky decomposition on a matrix
 *
 * This routine finds matrix L such that LL' = A. A must be positive definite.
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is.
 *
 *
 * Returns:
 *    Returns ERROR on failure, setting the the errors string accordingly, and 
 *    NO_ERROR otherwise.
 *
 * Author:
 *    Ernesto Brau
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/

int cholesky_decomposition(Matrix** L, const Matrix* A)
{
    Matrix* Lp = NULL;
    double sum;

    int i, j, k;

    if(A == NULL)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(get_zero_matrix(L, A->num_rows, A->num_cols));
    Lp = *L;

    for(i = 0; i < A->num_rows; i++)
    {
        for(j = 0; j <= i; j++)
        {
            if(i == j)
            {
                sum = 0;
                for(k = 0; k < i; k++)
                {
                    sum += Lp->elements[i][k] * Lp->elements[i][k];
                }

                if(A->elements[i][j] < sum)
                {
                    set_error("Cholesky decomposition: matrix is not positive definite.");
                    return ERROR;
                }
                Lp->elements[i][j] = sqrt(A->elements[i][j] - sum);
            }
            else
            {
                sum = 0;
                for(k = 0; k < j; k++)
                {
                    sum += Lp->elements[i][k] * Lp->elements[j][k];
                }
                Lp->elements[i][j] = (Lp->elements[j][j] != 0) ? (A->elements[i][j] - sum) / Lp->elements[j][j] : 0;
            }
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef __cplusplus
}
#endif

